#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#


"""This module contains a RegistryTree data model,
and a RegistryViewer dialog.
The puprose of those is to create a visualisation
of the classes that can be accessed from the
MDANSE IJob interface. 
"""

from MDANSE.Framework.Jobs.IJob import IJob

from qtpy.QtGui import QStandardItemModel, QStandardItem
from qtpy.QtCore import Slot, Signal, QSortFilterProxyModel, QModelIndex
from qtpy.QtWidgets import (
    QDialog,
    QTreeView,
    QGridLayout,
    QLabel,
    QApplication,
    QLineEdit,
    QTextEdit,
)
from qtpy.QtCore import Qt

# we check what attributes always exist in a Python object.
# we will not include those in the tree view, since they
# are not defined by the MDANSE code.
to_be_omitted = dir(None)


class RegistryTree(QStandardItemModel):
    """RegistryTree creates a tree structure
    of QStandardItem objects, and stores information
    about the names and docstrings of different
    classes contained in the IJob object.

    It inherits the QStandardItemModel, so it can be
    used in the Qt data/view/proxy model.
    """

    doc_string = Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._nodes = {}  # dict of {number: QStandardItem}
        self._docstrings = {}  # dict of {number: str}
        self._values = {}  # dict of {number: str}

        self._converters = {}
        self._jobs = {}

        self._by_ancestor = {}  # dict of list[int]

        self.nodecounter = 0  # each node is given a unique number
        # the number is the key to the dictionary entries.

        self.populateTree()
        # Some useful debugging here, for now
        # print("Converters")
        # print(self._converters)
        # print("Jobs")
        # print(self._jobs)

    def populateTree(self):
        """This function starts the recursive process of scanning
        the registry tree. Only called once on startup.
        """
        full_dict = IJob.indirect_subclass_dictionary()
        for key in full_dict.keys():
            self.parseNode(full_dict, key)

    def addNode(self, thing, name="Registry", parent: int = -1):
        """This function adds a new node to the tree data model.

        Arguments:
            thing -- a Python class from MDANSE.Framework

        Keyword Arguments:
            name -- the text tag that will be shown in the tree view
                    (default: {"Registry"})
            parent -- number (key) of the parent node in the data model
                    (default: {-1})

        Returns:
            the number of the new node
        """
        item = QStandardItem(name)
        retval = int(self.nodecounter)
        item.setData(retval, role=Qt.ItemDataRole.UserRole)
        self._nodes[self.nodecounter] = item
        self._docstrings[self.nodecounter] = thing.__doc__
        self.nodecounter += 1
        if parent < 0:
            self.appendRow(item)
        else:
            parent_node = self._nodes[parent]
            parent_node.appendRow(item)
        # here start the finer separation into categories
        if hasattr(thing, "ancestor") and hasattr(thing, "section"):
            try:
                cat = thing.category
            except AttributeError:
                pass
            else:
                if len(thing.ancestor) > 0:
                    for ancestor in thing.ancestor:
                        if ancestor not in self._by_ancestor.keys():
                            self._by_ancestor[ancestor] = []
                        self._by_ancestor[ancestor].append(retval)
                try:
                    is_conv = cat[0] == "Converters"
                except IndexError:
                    pass
                else:
                    if is_conv:
                        self._converters[name] = thing.__name__
                    else:
                        self._jobs[name] = thing.__name__
        # end of the finer assignment
        return retval

    def addTerminalNodes(self, thing, parent: int = -1):
        """A separate function for adding nodes to the
        tree structure which are _attributes_ and not
        _classes_.

        Arguments:
            thing -- the attribute of the parent node
            class to be stored here.

        Keyword Arguments:
            parent -- the number of the parent node (default: {-1})
        """
        for attr in dir(thing):
            if attr not in to_be_omitted:
                item = QStandardItem(str(attr))
                attr_object = getattr(thing, attr, None)
                item.setData(self.nodecounter, role=Qt.ItemDataRole.UserRole)
                # self._items.append(item)
                self._nodes[self.nodecounter] = item
                self._docstrings[self.nodecounter] = attr_object.__doc__
                self._values[self.nodecounter] = str(attr_object)
                self.nodecounter += 1
                if parent < 0:
                    self.appendRow(item)
                else:
                    parent_node = self._nodes[parent]
                    parent_node.appendRow(item)

    def parseNode(self, node, name="unknown", parent=-1):
        """A recursive function which discovers information
        about the Python object to be added to the data tree,
        and decides where to add it.

        Arguments:
            node -- the Python object to be added to the data
                   tree

        Keyword Arguments:
            name -- the name tag to be used for the new node
                   (default: {"unknown"})
            parent -- the number of the parent of the new node
                   (default: {-1})
        """
        parent_number = self.addNode(node, name, parent)
        if hasattr(node, "keys"):
            try:
                node.keys()
            except TypeError:
                self.addTerminalNodes(node, parent_number)
            else:
                for key in node.keys():
                    child = node[key]
                    self.parseNode(child, name=str(key), parent=parent_number)
        else:
            self.addTerminalNodes(node, parent_number)

    def getAncestry(self, number: int):
        """This function discovers the keys needed to be put
        into the REGISTRY to get the class attached to the
        specific node.

        Arguments:
            number -- number of the object in the data model.

        Returns:
            list[str] of all the parents of the given node.
        """
        parents = []
        item = self._nodes[number]
        while item is not None:
            parents.append(item.text())
            item = item.parent()
        return parents[::-1]


class RegistryViewer(QDialog):
    """A widget displaying the contents of the MDANSE
    REGISTRY.
    Normally intended to be used as a standalone
    program, since it is of little interest to the
    end users of MDANSE.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWindowTitle("MDANSE Registry Viewer")

        self.glayout = QGridLayout(self)
        self.setLayout(self.glayout)

        self.viewer = QTreeView(self)
        self.viewer.setObjectName("MDANSE Registry")
        self.glayout.addWidget(self.viewer, 0, 0, 4, 2)

        self.glayout.addWidget(QLabel("Filter", self), 4, 0)

        self.filter_field = QLineEdit(self)
        self.glayout.addWidget(self.filter_field, 4, 1)

        self.path_field = QLineEdit(self)
        self.glayout.addWidget(self.path_field, 0, 2)

        self.proxy = QSortFilterProxyModel(self)

        self.datamodel = RegistryTree(self)
        self.proxy.setSourceModel(self.datamodel)

        self.viewer.setModel(self.proxy)

        self.value_panel = QTextEdit(self)
        self.glayout.addWidget(self.value_panel, 1, 2, 2, 1)

        self.doc_panel = QTextEdit(self)
        self.glayout.addWidget(self.doc_panel, 3, 2, 1, 1)

        self.filter_field.textChanged.connect(self.filterEntries)

        self.viewer.clicked.connect(self.updateDocstring)
        self.viewer.clicked.connect(self.updatePath)

    @Slot()
    def filterEntries(self):
        phrase = self.filter_field.text()
        self.proxy.setFilterFixedString(phrase)

    @Slot(QModelIndex)
    def updateDocstring(self, index=QModelIndex):
        n_index = self.viewer.currentIndex()
        number = self.proxy.itemData(n_index)[256]
        try:
            docstring = self.datamodel._docstrings[number]
        except AttributeError:
            docstring = ""
        self.doc_panel.setText(docstring)
        try:
            valstring = self.datamodel._values[number]
        except KeyError:
            valstring = ""
        self.value_panel.setText(valstring)

    @Slot(QModelIndex)
    def updatePath(self, index=QModelIndex):
        n_index = self.viewer.currentIndex()
        number = self.proxy.itemData(n_index)[256]
        parents = self.datamodel.getAncestry(number)
        try:
            valstring = self.datamodel._values[number]
        except KeyError:
            pathstring = "REGISTRY['" + "']['".join(parents) + "']"
        else:
            pathstring = "REGISTRY['" + "']['".join(parents[:-1]) + "']." + parents[-1]
        self.path_field.setText(pathstring)


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    root = RegistryViewer()
    root.show()
    app.exec()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import TYPE_CHECKING, List
import math

import numpy as np
from scipy.interpolate import interp1d
from matplotlib.pyplot import colorbar as mpl_colorbar

from MDANSE.Framework.Units import measure
from MDANSE.MLogging import LOG

from MDANSE_GUI.Tabs.Plotters.Plotter import Plotter

if TYPE_CHECKING:
    from matplotlib.figure import Figure
    from MDANSE_GUI.Tabs.Models.PlottingContext import PlottingContext


class Heatmap(Plotter):

    def __init__(self) -> None:
        super().__init__()
        self._figure = None
        self._current_colours = []
        self._backup_images = {}
        self._backup_arrays = {}
        self._backup_minmax = {}
        self._backup_scale_interpolators = {}
        self._backup_limits = {}
        self._last_axes_units = {}
        self._initial_values = [0.0, 100.0]
        self._slider_values = [0.0, 100.0]
        self._last_minmax = [-1, -1]

    def clear(self, figure: "Figure" = None):
        if figure is None:
            target = self._figure
        else:
            target = figure
        if target is None:
            return
        target.clear()

    def slider_labels(self) -> List[str]:
        return ["Minimum (percentile)", "Maximum (percentile)"]

    def slider_limits(self) -> List[str]:
        return self._number_of_sliders * [[0.0, 100.0, 0.1]]

    def sliders_coupled(self) -> bool:
        return True

    def get_figure(self, figure: "Figure" = None):
        if figure is None:
            target = self._figure
        else:
            target = figure
        if target is None:
            LOG.error(f"PlottingContext can't plot to {target}")
            return
        target.clear()
        return target

    def handle_slider(self, new_value: List[float]):
        super().handle_slider(new_value)
        target = self._figure
        if target is None:
            return
        if new_value[1] <= new_value[0]:
            return
        else:
            self._slider_values = [new_value[0], new_value[1]]
        for ds_num, image in self._backup_images.items():
            try:
                last_minmax = self._backup_minmax[ds_num]
            except KeyError:
                self._backup_minmax[ds_num] = [-1, -1]
                last_minmax = [-1, -1]
            interpolator = self._backup_scale_interpolators[ds_num]
            newmax = interpolator(new_value[1])
            newmin = interpolator(new_value[0])
            if newmax < newmin:
                if newmax == last_minmax[1]:
                    newmin = float(newmax)
                else:
                    newmax = float(newmin)
            if newmin == last_minmax[0] and newmax == last_minmax[1]:
                return
            if newmax >= newmin:
                image.set_clim([newmin, newmax])
                self._figure.canvas.draw_idle()
                self._backup_minmax[ds_num] = [newmin, newmax]
        target.canvas.draw()

    def plot(
        self,
        plotting_context: "PlottingContext",
        figure: "Figure" = None,
        colours=None,
        update_only=False,
        toolbar=None,
    ):
        target = self.get_figure(figure)
        if target is None:
            return
        if toolbar is not None:
            self._toolbar = toolbar
        self._figure = target
        self._backup_images = {}
        self._backup_arrays = {}
        self._backup_scale_interpolators = {}
        xaxis_unit = None
        yaxis_unit = None
        self.get_mpl_colors()
        self._axes = []
        if not update_only:
            self._last_axes_units = {}
        self.apply_settings(plotting_context, colours)
        if plotting_context.set_axes() is None:
            LOG.error("Axis check failed.")
            return
        nplots = 0
        for databundle in plotting_context.datasets().values():
            ds, _, _, _, ds_num = databundle
            if ds._n_dim == 1:
                continue
            nplots += 1
            try:
                self._backup_scale_interpolators[ds_num](51.2)
            except:
                percentiles = np.linspace(0, 100.0, 21)
                results = [np.percentile(ds._data, perc) for perc in percentiles]
                self._backup_scale_interpolators[ds_num] = interp1d(
                    percentiles, results
                )
        gridsize = int(math.ceil(nplots**0.5))
        startnum = 1
        for num, databundle in enumerate(plotting_context.datasets().values()):
            dataset, _, _, _, ds_num = databundle
            if dataset._n_dim == 1:
                continue
            axes = target.add_subplot(gridsize, gridsize, startnum)
            startnum += 1
            self._axes.append(axes)
            limits = []
            axis_units = []
            for key, value in dataset._axes_units.items():
                target_unit = plotting_context.get_conversion_factor(value)
                axis_array = dataset._axes[key]
                try:
                    conversion_factor = measure(1.0, value, equivalent=True).toval(
                        target_unit
                    )
                except:
                    limits += [axis_array[0], axis_array[-1]]
                    axis_units.append(value)
                else:
                    limits += [
                        axis_array[0] * conversion_factor,
                        axis_array[-1] * conversion_factor,
                    ]
                    axis_units.append(target_unit)
            image = axes.imshow(
                dataset._data.T[::-1, :],
                extent=limits,
                aspect="auto",
                interpolation=None,
                cmap=plotting_context.colormap,
            )
            colorbar = mpl_colorbar(image, ax=image.axes, format="%.1e", pad=0.02)
            colorbar.set_label(dataset._data_unit)
            xlimits, ylimits = axes.get_xlim(), axes.get_ylim()
            self._backup_arrays[ds_num] = np.nan_to_num(dataset._data)
            if update_only:
                interpolator = self._backup_scale_interpolators[ds_num]
                last_minmax = [
                    interpolator(self._slider_values[0]),
                    interpolator(self._slider_values[1]),
                ]
                image.set_clim(last_minmax)
                try:
                    last_limits = self._backup_limits[ds_num]
                except KeyError:
                    self._backup_limits[ds_num] = [
                        xlimits[0],
                        xlimits[1],
                        ylimits[0],
                        ylimits[1],
                    ]
                    last_limits = self._backup_limits[ds_num]
                else:
                    try:
                        previous_units = self._last_axes_units[ds_num]
                    except KeyError:
                        self._last_axes_units[ds_num] = [axis_units[0], axis_units[1]]
                        previous_units = self._last_axes_units[ds_num]
                    if axis_units[0] == previous_units[0]:
                        axes.set_xlim((last_limits[0], last_limits[1]))
                    else:
                        self._last_axes_units[ds_num][0] = axis_units[0]
                        xlim = axes.get_xlim()
                        self._backup_limits[ds_num][0] = xlim[0]
                        self._backup_limits[ds_num][1] = xlim[1]
                    if axis_units[1] == previous_units[1]:
                        axes.set_ylim((last_limits[2], last_limits[3]))
                    else:
                        self._last_axes_units[ds_num][1] = axis_units[1]
                        ylim = axes.get_ylim()
                        self._backup_limits[ds_num][2] = ylim[0]
                        self._backup_limits[ds_num][3] = ylim[1]
            else:
                try:
                    last_limits = self._backup_limits[ds_num]
                except KeyError:
                    self._backup_limits[ds_num] = [
                        xlimits[0],
                        xlimits[1],
                        ylimits[0],
                        ylimits[1],
                    ]
                    last_limits = self._backup_limits[ds_num]
                interpolator = self._backup_scale_interpolators[ds_num]
                last_minmax = [
                    interpolator(self._slider_values[0]),
                    interpolator(self._slider_values[1]),
                ]
                image.set_clim(last_minmax)
                self._backup_minmax[ds_num] = [dataset._data.min(), dataset._data.max()]
                self._backup_limits[ds_num] = [
                    xlimits[0],
                    xlimits[1],
                    ylimits[0],
                    ylimits[1],
                ]
                self._last_axes_units[ds_num] = axis_units
            axes.set_xlabel(axis_units[0])
            axes.set_ylabel(axis_units[1])
            self._backup_images[ds_num] = image
        # axes.grid(True)
        self.request_slider_values()
        target.canvas.draw()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import TYPE_CHECKING, Dict, List
import os
import itertools

if TYPE_CHECKING:
    import h5py

import numpy as np
from matplotlib.markers import MarkerStyle
from matplotlib.lines import lineStyles
from matplotlib import rcParams
import matplotlib.pyplot as mpl
from qtpy.QtCore import Slot, Signal, QModelIndex, Qt
from qtpy.QtGui import QStandardItemModel, QStandardItem

from MDANSE.MLogging import LOG


def get_mpl_markers():
    unique_keys = list(set(MarkerStyle.markers.keys()))
    filtered_markers = {
        key: MarkerStyle.markers[key]
        for key in unique_keys
        if key not in ["", " ", "none"] + [str(x) for x in range(10)]
    }
    return filtered_markers


def get_mpl_lines():
    filtered_line_styles = {
        key: value for key, value in lineStyles.items() if key not in ["", " "]
    }
    return filtered_line_styles


def get_mpl_colours():
    cycler = rcParams["axes.prop_cycle"]
    colours = cycler.by_key()["color"]
    return colours


class SingleDataset:

    def __init__(self, name: str, source: "h5py.File"):
        self._name = name
        self._filename = source.filename
        self._curves = {}
        self._curve_labels = {}
        self._planes = {}
        bare_name = os.path.split(self._filename)[-1]
        self._labels = {
            "minimal": name,
            "medium": f"{bare_name}:{name}",
            "full": f"{self._filename}:{name}",
        }
        self._valid = True
        try:
            self._data = source[name][:]
        except KeyError:
            LOG.error(f"{name} is not a data set in the file")
            self._valid = False
            return
        except TypeError:
            self._valid = False
            LOG.error(f"{name} is not plottable")
            return
        self._data_unit = source[name].attrs["units"]
        self._n_dim = len(self._data.shape)
        self._axes_tag = source[name].attrs["axis"]
        self._axes = {}
        self._axes_units = {}
        for axis_name in self._axes_tag.split("|"):
            aname = axis_name.strip()
            if aname == "index":
                self._axes[aname] = np.arange(len(self._data))
                self._axes_units[aname] = "N/A"
            else:
                self._axes[aname] = source[aname][:]
                self._axes_units[aname] = source[aname].attrs["units"]

    def available_x_axes(self) -> List[str]:
        return list(self._axes_units.values())

    def longest_axis(self) -> str:
        length = -1
        best_unit = "none"
        best_axis = "none"
        for aname, aunit in self._axes_units.items():
            xaxis = self._axes[aname]
            xlen = len(xaxis)
            if xlen > length:
                length = xlen
                best_unit = aunit
                best_axis = aname
        return best_unit, best_axis

    def curves_vs_axis(self, axis_unit: str) -> List[np.ndarray]:
        found = -1
        total_ndim = len(self._data.shape)
        data_shape = self._data.shape
        for aname, aunit in self._axes_units.items():
            if aunit == axis_unit:
                xaxis = self._axes[aname]
                xlen = len(xaxis)
                for dim in range(total_ndim):
                    if xlen == self._data.shape[dim]:
                        found = dim
                        break
        slicer = []
        indexer = []
        for dim in range(total_ndim):
            if dim == found:
                slicer.append([slice(None)])
                # indexer.append([None])
            else:
                slicer.append(np.arange(data_shape[dim]))
                indexer.append(np.arange(data_shape[dim]))
        indices = list(itertools.product(*indexer))
        slicers = list(itertools.product(*slicer))
        for n in range(len(indices)):
            self._curves[tuple(indices[n])] = self._data[slicers[n]].squeeze()
            self._curve_labels[tuple(indices[n])] = str(tuple(indices[n]))
        return self._curves
        # slicer = tuple(slicer)
        # temp = self._data[slicer].squeeze()
        # for line in temp:
        #     if len(line) != xlen:
        #         print("Wrong data length in the curves_vs_axis method of PlottingContext")
        # return temp


class SingleCurve:

    def __init__(self, data_name: str, file_name: str, *args, **kwargs):
        self._name = data_name
        self._filename = file_name
        self._curve = []
        self._x_axis = []
        self._y_axis = None
        self._z_axis = None
        bare_name = os.path.split(self._filename)[-1]
        self._labels = {
            "minimal": data_name,
            "medium": f"{bare_name}:{data_name}",
            "full": f"{file_name}:{data_name}",
        }
        self._label = self._labels["medium"]
        self._units = {"data": "arbitrary", "x": None, "y": None, "z": None}
        self._colour = "#000000"
        self._line = "-"

    def set_data(self, array: np.ndarray, unit: str):
        self._curve = array
        self._units["data"] = unit

    def set_x_axis(self, array: np.ndarray, unit: str):
        self._x_axis = array
        self._units["x"] = unit

    def set_y_axis(self, value: float, unit: str):
        self._y_axis = value
        self._units["y"] = unit

    def set_z_axis(self, value: float, unit: str):
        self._z_axis = value
        self._units["z"] = unit

    def standard_items(self, key: int) -> List["QStandardItem"]:
        result = []
        for val in []:
            item = QStandardItem(str(val))
            item.setData(key, role=Qt.ItemDataRole.UserRole)
            result.append(item)
        return result


class PlottingContext(QStandardItemModel):

    needs_an_update = Signal()

    def __init__(self, *args, unit_lookup=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._datasets = {}
        self._current_axis = [None, None, None]
        self._figure = None
        self._ndim_lowest = 1
        self._ndim_highest = 3
        self._all_xunits = []
        self._best_xunits = []
        self._colour_list = get_mpl_colours()
        self._last_colour_list = get_mpl_colours()
        self._colour_map = kwargs.get("colormap", "viridis")
        self._last_colour = 0
        self._unit_lookup = unit_lookup
        self.setHorizontalHeaderLabels(
            [
                "Dataset",
                "Trajectory",
                "Size",
                "Unit",
                "Use it?",
                "Colour",
                "Line style",
                "Marker",
            ]
        )

    def generate_colour(self, number: int):
        return self._colour_list[number % len(self._colour_list)]

    def next_colour(self):
        colour = self.generate_colour(self._last_colour)
        self._last_colour += 1
        return colour

    @property
    def colormap(self):
        backup_cmap = "viridis"
        try:
            cmap = self._unit_lookup._settings.group("colours").get("colormap")
        except:
            return backup_cmap
        else:
            if cmap in mpl.colormaps():
                return cmap
            else:
                return self._unit_lookup._settings.default_value("colours", "colormap")

    @Slot()
    def regenerate_colours(self):
        self._colour_list = get_mpl_colours()
        self._last_colour = 0
        for row in range(self.rowCount()):
            current_colour = self.item(row, 5).text()
            if (
                current_colour
                != self._last_colour_list[row % len(self._last_colour_list)]
            ):
                self._last_colour += 1
            else:
                self.item(row, 5).setText(str(self.next_colour()))

    @Slot(object)
    def accept_external_data(self, other: "PlottingContext"):
        """Crucial slot for transferring data
        between tabs. DataPlotter will send the datasets
        into this slot."""
        for dataset in other._datasets.values():
            self.add_dataset(dataset)
        self.set_axes()

    @Slot(dict)
    def accept_units(self, units: Dict):
        """Crucial slot for transferring data
        between tabs. DataPlotter will send the datasets
        into this slot."""
        self._unit_preference = units
        self.set_axes()

    def get_conversion_factor(self, unit):
        factor, new_unit = self._unit_lookup.conversion_factor(unit)
        return new_unit

    def datasets(self) -> Dict:
        result = {}
        for ds_num, row in enumerate(range(self.rowCount())):
            key = self.index(row, 0).data(role=Qt.ItemDataRole.UserRole)
            useit = (
                self.itemFromIndex(self.index(row, 4)).checkState()
                == Qt.CheckState.Checked
            )
            colour = self.itemFromIndex(self.index(row, 5)).text()
            style = self.itemFromIndex(self.index(row, 6)).text()
            marker = self.itemFromIndex(self.index(row, 7)).text()
            if useit:
                result[key] = (self._datasets[key], colour, style, marker, ds_num)
        return result

    def add_dataset(self, new_dataset: SingleDataset):
        if not new_dataset._valid:
            return
        newkey = f"{new_dataset._filename}:{new_dataset._name}"
        if newkey in self._datasets:
            return
        self._datasets[newkey] = new_dataset
        items = [
            QStandardItem(str(x))
            for x in [
                new_dataset._name,
                new_dataset._filename,
                new_dataset._data.shape,
                new_dataset._data_unit,
                "",
                self.next_colour(),
                "-",
                "",
            ]
        ]
        for item in items:
            item.setData(newkey, role=Qt.ItemDataRole.UserRole)
        for item in items[:5]:
            item.setEditable(False)
        temp = items[4]
        temp.setCheckable(True)
        temp.setCheckState(Qt.CheckState.Checked)
        self.itemChanged.connect(self.needs_an_update)
        # test for possible nested items
        best_axis = new_dataset.longest_axis()
        curves = new_dataset.curves_vs_axis(best_axis[0])
        if len(curves) > 1:
            counter = 1
            for indices, curve in curves.items():
                temp = SingleCurve(new_dataset._name, new_dataset._filename)
                temp.set_data(curve, new_dataset._data_unit)
                temp.set_x_axis(new_dataset._axes[best_axis[1]], best_axis[0])
                # optionally, we should set y and z as well
                subitems = temp.standard_items(newkey + f":{counter}")
                items[0].appendRow(subitems)
                counter += 1
        else:
            LOG.error("No curves!")
        self.appendRow(items)

    def set_axes(self):
        if len(self._datasets) == 0:
            return
        all_axes = []
        all_dimensions = []
        longest_axes = []
        unit_to_axname = {}
        for dataset in self._datasets.values():
            n_dimensions = dataset._n_dim
            if n_dimensions not in all_dimensions:
                all_dimensions.append(n_dimensions)
            best_units, best_axis = dataset.longest_axis()
            unit_to_axname[best_units] = best_axis
            for unit in dataset._axes_units.values():
                if unit not in all_axes:
                    all_axes.append(unit)
            if best_units not in longest_axes:
                longest_axes.append(best_units)
        self._ndim = 1
        self._ndim_highest = np.max(all_dimensions)
        self._ndim_lowest = np.min(all_dimensions)
        self._best_xunits = longest_axes
        self._all_xunits = all_axes
        self._unit_to_axname = unit_to_axname
        # self.needs_an_update.emit()
        return "Configured!"

    @Slot()
    def clear(self) -> None:
        result = super().clear()
        self._datasets = {}
        return result

    @Slot(QModelIndex)
    def delete_dataset(self, index: QModelIndex):
        dkey = index.data(role=Qt.ItemDataRole.UserRole)
        self.removeRow(index.row())
        self._datasets.pop(dkey, None)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtCore import qInstallMessageHandler
from qtpy.QtWidgets import QWidget

from MDANSE_GUI.Tabs.GeneralTab import GeneralTab
from MDANSE_GUI.Tabs.Layouts.SinglePanel import SinglePanel
from MDANSE_GUI.Session.LocalSession import LocalSession
from MDANSE_GUI.Tabs.Visualisers.TextInfo import TextInfo


log_tab_label = """MDANSE_GUI message log.
"""


class LoggingTab(GeneralTab):
    """The tab for tracking the progress of running jobs."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        qInstallMessageHandler(self.log_handler)

    def log_handler(self, m_type, m_context, m_text):
        self._visualiser.append_text(m_text)

    @classmethod
    def standard_instance(cls):
        the_tab = cls(
            window,
            name="Logger",
            visualiser=TextInfo(),
            layout=SinglePanel,
            label_text=log_tab_label,
        )
        return the_tab

    @classmethod
    def gui_instance(
        cls,
        parent: QWidget,
        name: str,
        session: LocalSession,
        settings,
        logger,
        **kwargs,
    ):
        the_tab = cls(
            parent,
            name=name,
            session=session,
            settings=settings,
            logger=logger,
            visualiser=TextInfo(),
            layout=SinglePanel,
            label_text=log_tab_label,
        )
        return the_tab


if __name__ == "__main__":
    import sys
    from qtpy.QtWidgets import QApplication, QMainWindow, QVBoxLayout

    app = QApplication(sys.argv)
    window = QMainWindow()
    the_tab = LoggingTab(
        window,
        name="RunningJobs",
        visualiser=TextInfo(),
        layout=SinglePanel,
        label_text=log_tab_label,
    )
    window.setCentralWidget(the_tab._core)
    window.show()
    app.exec()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import vtk

RGB_COLOURS = []
RGB_COLOURS.append((1.00, 0.20, 1.00))  # selection
RGB_COLOURS.append((1.00, 0.90, 0.90))  # background


class ColourManager:
    def __init__(self, *args, init_colours: list = None, **kwargs):
        self._lut = vtk.vtkColorTransferFunction()
        if init_colours is None:
            self._colour_list = RGB_COLOURS
        else:
            self._colour_list = init_colours
        self.rebuild_colours()

    def clear_table(self):
        """This was meant to be used for cleaning up,
        but the underlying VTK object seems to survive.
        The unit test shows 5 colours in the list when there
        should be 2.
        """
        self._lut.RemoveAllPoints()

    def rebuild_colours(self):
        """Puts the colours from the internal _colour_list
        into the VTK vtkColorTransferFunction object.
        The colours stored internally in self_.colour_list
        should be the same as the colours in self._lut
        AFTER this method has been called.
        """
        self._lut.RemoveAllPoints()

        for index, colour in enumerate(self._colour_list):
            self._lut.AddRGBPoint(index, colour[0], colour[1], colour[2])

    def add_colour(self, rgb: tuple[int, int, int]) -> int:
        """Puts the colour into the list and returns the
        index of the colour in the list, to be used by VTK.

        Arguments:
            rgb -- RGB colour values for the colour

        Returns:
            int - the index of the colour in the list
        """
        float_rgb = tuple([x / 255.0 for x in rgb])
        for index, value in enumerate(self._colour_list):
            if value == float_rgb:
                return index
        new_index = len(self._colour_list)
        self._colour_list.append(float_rgb)
        self.rebuild_colours()
        return new_index

    def initialise_from_database(
        self, atoms: list[str], element_database=None
    ) -> list[int]:
        """Puts colours into the list based on chemical elements list
        and a database with colour values.

        Arguments:
            atoms -- list[str] of chemical element names

        Keyword Arguments:
            element_database -- a dictionary containing RGB values for chemical elements,
               typically the MDANSE_GUI.MolecularViewer.database.CHEMICAL_ELEMENTS

        Returns:
            list[int] -- a list of indices of colours, with one numbed per atom.
        """
        index_list = []
        for atom in atoms:
            rgb = (int(x) for x in element_database["atoms"][atom]["color"].split(";"))
            index_list.append(self.add_colour(rgb))
        return index_list



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#


"""This module contains a RegistryTree data model,
and a RegistryViewer dialog.
The puprose of those is to create a visualisation
of the classes that can be accessed from the
MDANSE IJob interface. 
"""


from qtpy.QtCore import Slot, QSortFilterProxyModel, QModelIndex
from qtpy.QtWidgets import (
    QDialog,
    QTreeView,
    QGridLayout,
    QLabel,
    QApplication,
    QLineEdit,
    QTextEdit,
)

# we check what attributes always exist in a Python object.
# we will not include those in the tree view, since they
# are not defined by the MDANSE code.
to_be_omitted = dir(None)


class RegistryViewer(QDialog):
    """A widget displaying the contents of the MDANSE
    REGISTRY.
    Normally intended to be used as a standalone
    program, since it is of little interest to the
    end users of MDANSE.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWindowTitle("MDANSE Registry Viewer")

        self.glayout = QGridLayout(self)
        self.setLayout(self.glayout)

        self.viewer = QTreeView(self)
        self.viewer.setObjectName("MDANSE Registry")
        self.glayout.addWidget(self.viewer, 0, 0, 4, 2)

        self.glayout.addWidget(QLabel("Filter", self), 4, 0)

        self.filter_field = QLineEdit(self)
        self.glayout.addWidget(self.filter_field, 4, 1)

        self.path_field = QLineEdit(self)
        self.glayout.addWidget(self.path_field, 0, 2)

        self.proxy = QSortFilterProxyModel(self)

        self.datamodel = RegistryTree(self)
        self.proxy.setSourceModel(self.datamodel)

        self.viewer.setModel(self.proxy)

        self.value_panel = QTextEdit(self)
        self.glayout.addWidget(self.value_panel, 1, 2, 2, 1)

        self.doc_panel = QTextEdit(self)
        self.glayout.addWidget(self.doc_panel, 3, 2, 1, 1)

        self.filter_field.textChanged.connect(self.filterEntries)

        self.viewer.clicked.connect(self.updateDocstring)
        self.viewer.clicked.connect(self.updatePath)

    @Slot()
    def filterEntries(self):
        phrase = self.filter_field.text()
        self.proxy.setFilterFixedString(phrase)

    @Slot(QModelIndex)
    def updateDocstring(self, index=QModelIndex):
        n_index = self.viewer.currentIndex()
        number = self.proxy.itemData(n_index)[256]
        try:
            docstring = self.datamodel._docstrings[number]
        except AttributeError:
            docstring = ""
        self.doc_panel.setText(docstring)
        try:
            valstring = self.datamodel._values[number]
        except KeyError:
            valstring = ""
        self.value_panel.setText(valstring)

    @Slot(QModelIndex)
    def updatePath(self, index=QModelIndex):
        n_index = self.viewer.currentIndex()
        number = self.proxy.itemData(n_index)[256]
        parents = self.datamodel.getAncestry(number)
        try:
            valstring = self.datamodel._values[number]
        except KeyError:
            pathstring = "REGISTRY['" + "']['".join(parents) + "']"
        else:
            pathstring = "REGISTRY['" + "']['".join(parents[:-1]) + "']." + parents[-1]
        self.path_field.setText(pathstring)


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    root = RegistryViewer()
    root.show()
    app.exec()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from qtpy.QtCore import Slot, Signal, QObject, QTimer
from qtpy.QtWidgets import QProgressBar


class FlexBar(QProgressBar):
    def __init__(self, parent):
        super().__init__(parent)
        self.forward = True
        self.setTextVisible(False)

    @Slot()
    def reverse(self):
        if self.forward == True:
            self.forward = False
        else:
            self.forward = True
        self.setInvertedAppearance(not self.forward)


class Oscillator(QObject):
    nextvalue = Signal(int)
    turnaround = Signal()
    clearbar = Signal()

    def __init__(self, progbar, parent=None, totalsteps=100):
        super().__init__(parent)
        self.progbar = progbar
        self.progbar.setMaximum(totalsteps)
        self.currval = 0
        self.maxval = totalsteps
        self.minval = 0
        self.forward = True
        self.timer = QTimer(self)
        self.nextvalue.connect(progbar.setValue)
        self.turnaround.connect(progbar.reverse)
        self.clearbar.connect(progbar.reset)
        self.timer.setInterval(10)
        self.timer.timeout.connect(self.step)

    def start(self):
        self.timer.start()

    def stop(self):
        self.timer.stop()
        self.clearbar.emit()
        self.currval = 0
        self.forward = True
        self.progbar.setInvertedAppearance(False)

    @Slot()
    def step(self):
        if self.forward:
            self.currval += 1
        else:
            self.currval -= 1
        if self.currval >= self.maxval:
            self.forward = not self.forward
            self.turnaround.emit()
        elif self.currval <= self.minval:
            self.forward = not self.forward
        self.nextvalue.emit(self.currval)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
import typing

from qtpy.QtWidgets import QTreeView, QWidget, QMenu
from qtpy.QtCore import Signal, Slot, QModelIndex
from qtpy.QtGui import QContextMenuEvent

from MDANSE.MLogging import LOG

from MDANSE_GUI.DataViewModel.TrajectoryHolder import DataTreeItem


class TrajectoryViewer(QTreeView):
    """A modified QTreeView which enables context menus to pop
    up for different elements of the workflow.
    """

    itemPicked = Signal(DataTreeItem)
    pickedAncestor = Signal(str)

    def __init__(self, parent: typing.Optional[QWidget] = None) -> None:
        super().__init__(parent)

        self.clicked.connect(self.requestChildren)

    def requestChildren(self, index: QModelIndex):
        # item = index.data()
        item = self.model().itemFromIndex(index)
        LOG.info(f"Emitting items ancestor. Item: {item}")
        anc = item.ancestors()
        LOG.info(f"ancestor: {anc[-1]}")
        # self.pickedAncestor.emit(anc[-1])
        self.itemPicked.emit(item)

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:
        index = self.indexAt(event.pos())
        model = self.model()
        item = model.itemData(index)
        # item = index.data()
        LOG.info(index)
        LOG.info(item)
        menu = QMenu()
        self.populateMenu(menu, item)
        # picked = menu.exec(menu.)
        menu.exec_(event.globalPos())
        # return super().contextMenuEvent(event)

    def populateMenu(self, menu: QMenu, item: DataTreeItem):
        for action, method in [("Delete", self.deleteNode)]:
            temp_action = menu.addAction(action)
            temp_action.triggered.connect(method)

    @Slot()
    def deleteNode(self):
        model = self.model()
        index = self.currentIndex()
        model.removeRow(index.row())



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtGui import QStandardItemModel, QStandardItem
from qtpy.QtCore import QObject, Slot, Signal, QProcess, QThread, QMutex

from MDANSE.MLogging import LOG

from MDANSE_GUI.DataViewModel.JobStatusQt import JobStatusQt


class JobThread(QThread):
    """A wrapper object for a single MDANSE analysis job.
    It is created to run a single instance of the job,
    and exits once the job has completed processing."""

    job_failure = Signal(str)

    def __init__(self, *args, command=None, parameters={}):
        super().__init__(*args)
        LOG.info("JobThread starts init")
        self._command = command
        self._parameters = parameters
        LOG.info(f"JobThread.run will create a job instance of command {command}")
        if isinstance(self._command, type):
            self._job = self._command()
        else:
            self._job = self._command
        self._job.build_configuration()
        LOG.info(f"JobThread._parameters: {self._parameters}")
        # here we try to create and connect a JobStatusQt
        status = JobStatusQt(parent=self)
        self._job._status = status
        self._status = status
        LOG.info("JobThread finished init")

    def run(self):
        try:
            self._job.run(self._parameters)
        except Exception as inst:
            LOG.error("JobThread has entered exception handling!")
            error_message = ""
            error_message += str(type(inst))
            error_message += str(inst.args)  # arguments stored in .args
            error_message += str(inst)  # __str__ allows args to be printed directly,
            LOG.error("JobThread is about to emit the failure message")
            self.job_failure.emit(error_message)
        else:
            LOG.info(
                "JobThread.run did not raise an exception. JobThread.run will exit now"
            )
        self.exec()  # this starts event handling - will it help?


class JobEntry(QObject):
    """This coordinates all the objects that make up one line on the list
    of current jobs. It is used for reporting the task progress to the GUI."""

    def __init__(self, *args, command=None):
        super().__init__(*args)
        self._command = command
        self._parameters = {}
        self.has_started = False
        self.has_finished = False
        self.success = None
        self.percent_complete = 0
        self._output = None
        self.reference = None
        self.total_steps = 99
        self._prog_item = QStandardItem()
        self._stat_item = QStandardItem()

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def set_parameters(self, input: dict):
        self._parameters = input

    def update_fields(self):
        self._prog_item.setText(f"{self.percent_complete} percent complete")

    @Slot(bool)
    def on_finished(self, success: bool):
        LOG.info("Item received on_finished!")
        self.success = success
        self.has_finished = True
        self._stat_item.setText("Stopped")
        if success:
            self.percent_complete = 100
            self._stat_item.setText("Completed!")
        self.update_fields()

    @Slot(int)
    def on_started(self, target_steps: int):
        LOG.info("Item received on_started!")
        self.total_steps = target_steps
        self.has_started = True
        self._stat_item.setText("Starting")
        self.update_fields()

    @Slot(int)
    def on_update(self, completed_steps: int):
        LOG.info("Item received on_update!")
        self.percent_complete = completed_steps / self.total_steps * 99
        self._stat_item.setText("Running")
        self.update_fields()

    @Slot()
    def on_oscillate(self):
        """For jobs with unknown duration, the progress bar will bounce."""


class JobHolder(QStandardItemModel):
    """All the job INSTANCES that are started by the GUI
    are added to this model."""

    def __init__(self, parent: QObject = None):
        super().__init__(parent=parent)
        self.lock = QMutex()
        self.existing_threads = []
        self.existing_jobs = []

    @Slot(str)
    def reportError(self, err: str):
        LOG.error(err)

    @Slot(object)
    def addItem(self, new_entry: QProcess):
        traj = JobEntry(new_entry.basename, trajectory=new_entry)
        self.appendRow([traj])

    @Slot(list)
    def startThread(self, job_vars: list):
        try:
            th_ref = JobThread(command=job_vars[0], parameters=job_vars[1])
        except:
            LOG.error(f"Failed to create JobThread using {job_vars}")
            return
        th_ref.job_failure.connect(self.reportError)
        item_th = JobEntry(command=job_vars[0])
        th_ref._status._communicator.target.connect(item_th.on_started)  # int
        th_ref._status._communicator.progress.connect(item_th.on_update)  # int
        th_ref._status._communicator.finished.connect(item_th.on_finished)  # bool
        th_ref._status._communicator.oscillate.connect(item_th.on_oscillate)  # nothing
        LOG.info("Thread ready to start!")
        try:
            task_name = str(job_vars[0].__name__)
        except AttributeError:
            task_name = str(job_vars[0].__class__.__name__)
        self.appendRow(
            [
                QStandardItem(task_name),
                item_th._prog_item,
                item_th._stat_item,
            ]
        )
        # nrows = self.rowCount()
        # index = self.indexFromItem(item_th._item)
        # print(f"Index: {index}")
        th_ref.start()
        self.existing_threads.append(th_ref)
        self.existing_jobs.append(item_th)



import pytest
import tempfile
from importlib import reload

import vtk
import numpy as np

from MDANSE_GUI.MolecularViewer.ColourManager import ColourManager, RGB_COLOURS
from MDANSE_GUI.MolecularViewer.database import CHEMICAL_ELEMENTS


@pytest.fixture(scope="function", params=[None, RGB_COLOURS])
def colour_manager(request):
    temp = ColourManager(init_colours=request.param)
    yield temp
    temp.clear_table()
    reload(vtk)


@pytest.mark.xfail(reason="see docstring")
def test_ColourList(colour_manager: ColourManager):
    """It seems that in the second run of this test
    the vtk object storing the information about colours is reused
    and still holds 5 elements, even though the Python object
    connected to it has been destroyed an initialised again.
    Until a way has been found to reset the vtk context,
    this test will not pass.

    Parameters
    ----------
    colour_manager : ColourManager
        A test fixture: ColourManager instance
    """
    lut_size = colour_manager._lut.GetSize()
    assert lut_size == 2  # we initialised with 2 colours
    indices = colour_manager.initialise_from_database(
        ["O", "C", "Fe"], CHEMICAL_ELEMENTS
    )
    assert np.all(np.array(indices) == [2, 3, 4])
    lut_size = colour_manager._lut.GetSize()
    assert lut_size == 5  # we added 3 colours, so it should be 5 now.



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtWidgets import (
    QDialog,
    QPushButton,
    QVBoxLayout,
    QWidget,
    QHBoxLayout,
)
from qtpy.QtCore import Signal, Slot

from MDANSE.MLogging import LOG
from MDANSE.Framework.Jobs.IJob import IJob

from MDANSE_GUI.Widgets.GeneralWidgets import InputFactory


class ConverterDialog(QDialog):
    new_thread_objects = Signal(list)
    new_path = Signal(str)

    def __init__(self, *args, converter: IJob = "Dummy", **kwargs):
        super().__init__(*args, **kwargs)

        layout = QVBoxLayout(self)
        self.setLayout(layout)
        self.handlers = {}
        self.converter_instance = None
        self.converter_constructor = converter
        self.default_path = "."

        converter_instance = converter
        converter_instance.build_configuration()
        settings = converter_instance.settings
        self.converter_instance = converter_instance
        for key, value in settings.items():
            dtype = value[0]
            ddict = value[1]
            defaultvalue = ddict.get("default", 0.0)
            labeltext = ddict.get("label", "Mystery X the Unknown")
            ddict["label"] = labeltext
            base, data_handler = InputFactory.createInputField(
                parent=self, kind=dtype, **ddict
            )
            layout.addWidget(base)
            self.handlers[key] = data_handler

        buttonbase = QWidget(self)
        buttonlayout = QHBoxLayout(buttonbase)
        buttonbase.setLayout(buttonlayout)
        self.cancel_button = QPushButton("Cancel", buttonbase)
        self.execute_button = QPushButton("CONVERT!", buttonbase)
        self.execute_button.setStyleSheet("font-weight: bold")

        self.cancel_button.clicked.connect(self.cancel_dialog)
        self.execute_button.clicked.connect(self.execute_converter)

        buttonlayout.addWidget(self.cancel_button)
        buttonlayout.addWidget(self.execute_button)

        layout.addWidget(buttonbase)

    @Slot(dict)
    def parse_updated_params(self, new_params: dict):
        if "path" in new_params.keys():
            self.default_path = new_params["path"]
            self.new_path.emit(self.default_path)

    @Slot()
    def cancel_dialog(self):
        self.destroy()

    @Slot()
    def execute_converter(self):
        if self.converter_instance is None:
            LOG.warning("No converter instance attached to the Dialog")
            return False
        pardict = {}
        LOG.info(f"handlers: {self.handlers}")
        for key, value in self.handlers.items():
            pardict[key] = value.returnValue()
        LOG.info(
            f"Passing {pardict} to the converter instance {self.converter_instance}"
        )
        self.new_thread_objects.emit([self.converter_constructor, pardict])



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from qtpy.QtWidgets import QTreeView, QComboBox, QItemDelegate, QColorDialog
from qtpy.QtCore import Signal, Slot, Qt
from qtpy.QtGui import QColor

from MDANSE_GUI.Tabs.Models.PlottingContext import get_mpl_lines, get_mpl_markers


class ColourPicker(QItemDelegate):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def createEditor(self, parent, option, index):
        dialog = QColorDialog(parent)
        return dialog

    def setEditorData(self, editor, index):
        colour_string = index.data()
        color = QColor(colour_string)
        editor.setCurrentColor(color)

    def setModelData(self, editor, model, index):
        if editor.result() == QColorDialog.DialogCode.Accepted:
            color = editor.currentColor()
            colour_string = color.toRgb()
            model.setData(index, colour_string)


class MplStyleCombo(QItemDelegate):

    def __init__(self, *args, mpl_items=None, **kwargs) -> None:
        self._items = list([str(x) for x in mpl_items.keys()])
        super().__init__(*args, **kwargs)

    def setEditorData(self, editor, index):
        editor.blockSignals(True)
        text = index.model().data(index, Qt.DisplayRole)
        try:
            i = self._items.index(text)
        except ValueError:
            i = 0
        editor.setCurrentIndex(i)
        editor.blockSignals(False)

    def setModelData(self, editor, model, index):
        model.setData(index, editor.currentText())

    def createEditor(self, parent, option, index):
        combo = QComboBox(parent)
        li = []
        for item in self._items:
            li.append(item)
        combo.addItems(li)
        combo.currentIndexChanged.connect(self.currentIndexChanged)
        return combo

    @Slot()
    def currentIndexChanged(self):
        self.commitData.emit(self.sender())


class PlotDetailsView(QTreeView):
    details_changed = Signal(object)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._delegates = {
            "colour": ColourPicker(),
            "line": MplStyleCombo(mpl_items=get_mpl_lines()),
            "marker": MplStyleCombo(mpl_items=get_mpl_markers()),
        }
        self.setItemDelegateForColumn(5, self._delegates["colour"])
        self.setItemDelegateForColumn(6, self._delegates["line"])
        self.setItemDelegateForColumn(7, self._delegates["marker"])

    def connect_to_visualiser(self, visualiser) -> None:
        """Connect to a visualiser.

        Parameters
        ----------
        visualiser : Action or TextInfo
            A visualiser to connect to this view.
        """
        self.details_changed.connect(visualiser.update_plot_details)



import fnmatch
import os
import sys

import numpy

from setuptools import setup, find_packages
from distutils.sysconfig import get_config_vars
from distutils.util import convert_path

try:
    import sphinx
except ImportError:
    sphinx = None


#################################
# Modules variables
#################################
EXCLUDE = ['*.py', '*.pyc', '*$py.class', '*~', '.*', '*.bak', '*.so', '*.pyd']

EXCLUDE_DIRECTORIES = ('.*', 'CVS', '_darcs', './build','*svn','./dist', 'EGG-INFO', '*.egg-info')

EXTENSIONS_PATH = "Extensions"

INCLUDE_DIR = [numpy.get_include()]

QHULL_DIR = os.path.join("Extensions","qhull_lib")

QHULL_INCLUDE_DIR = INCLUDE_DIR + [EXTENSIONS_PATH] + [os.path.join(QHULL_DIR,"ext")] + [os.path.join(QHULL_DIR,"src")]

#################################
# Helper function
#################################


def is_package(path):
    return (os.path.isdir(path) and os.path.isfile(os.path.join(path, '__init__.py')))


def find_package_data(where='.', package='', exclude=EXCLUDE, exclude_directories=EXCLUDE_DIRECTORIES, only_in_packages=True, show_ignored=False):
    
    out = {}
    stack = [(convert_path(where), '', package, only_in_packages)]
    while stack:
        where, prefix, package, only_in_packages = stack.pop(0)
        for name in os.listdir(where):
            fn = os.path.join(where, name)
            if os.path.isdir(fn):
                bad_name = False
                for pattern in exclude_directories:
                    if (fnmatch.fnmatchcase(name, pattern)
                        or fn.lower() == pattern.lower()):
                        bad_name = True
                        if show_ignored:
                            print >> sys.stderr, ("Directory %s ignored by pattern %s" % (fn, pattern))
                        break
                if bad_name:
                    continue
                if (os.path.isfile(os.path.join(fn, '__init__.py')) and not prefix):
                    if not package:
                        new_package = name
                    else:
                        new_package = package + '.' + name
                    stack.append((fn, '', new_package, False))
                else:
                    stack.append((fn, prefix + name + '/', package, only_in_packages))
            elif package or not only_in_packages:
                # is a file
                bad_name = False
                for pattern in exclude:
                    if (fnmatch.fnmatchcase(name, pattern)
                        or fn.lower() == pattern.lower()):
                        bad_name = True
                        if show_ignored:
                            print >> sys.stderr, ("File %s ignored by pattern %s" % (fn, pattern))
                        break
                if bad_name:
                    continue
                out.setdefault(package, []).append(prefix+name)
    
    return out


def find_data(where=".", exclude=EXCLUDE, exclude_directories=EXCLUDE_DIRECTORIES, prefix=""):

    out = {}
    stack = [convert_path(where)]
    while stack:
        where = stack.pop(0)
        for name in os.listdir(where):
            fn = os.path.join(where, name)
            d = os.path.join(prefix,os.path.dirname(fn))
            if os.path.isdir(fn):
                stack.append(fn)
            else:
                bad_name = False
                for pattern in exclude:
                    if (fnmatch.fnmatchcase(name, pattern) or fn.lower() == pattern.lower()):
                        bad_name = True
                        break
                if bad_name:
                    continue
                out.setdefault(d, []).append(fn)
            
    out = [(k,v) for k, v in out.items()]
    
    return out


#################################
# Documentation
#################################

if sphinx:

    try:
        from sphinx.ext.apidoc import main as sphinx_apidoc_main
    except ImportError:
        from sphinx.apidoc import main as sphinx_apidoc_main

#     class mdanse_build_doc(sphinx.setup_command.BuildDoc):
    class mdanse_build_doc():

        def run(self):
            
            build = self.get_finalized_command('build')

            buildDir = os.path.abspath(build.build_lib)

            if not os.path.exists(buildDir):
                raise IOError("build command must be performed prior building the doc")

            sys.path.insert(0,buildDir)

            sphinxDir = os.path.abspath(os.path.join(build.build_base,'sphinx',self.doctype))

            if not os.path.exists(sphinxDir):                
                os.makedirs(sphinxDir)
                                 
            metadata = self.distribution.metadata
            args = ["-F",
                    "--separate",
                    "-H%s" % metadata.name,
                    "-A%s" % metadata.author,
                    "-R%s" % metadata.version,
                    "-o%s" % sphinxDir,
                    os.path.join(buildDir,'MDANSE'),
                    os.path.join(buildDir,'MDANSE','Externals')]


            # /!\ apidoc.main is deprecated. The API has been broken in sphinx 1.7.0, see https://github.com/sphinx-doc/sphinx/issues/4615
            if int(sphinx.__version__.split(".")[1]) <= 6:
                args.insert(0,"")

            sphinx_apidoc_main(args)
                             
            currentDirectory = os.getcwd()
            
            import shutil
            shutil.copy(os.path.join(currentDirectory,'Doc','conf_%s.py' % self.doctype),os.path.join(sphinxDir,'conf.py'))
            shutil.copy(os.path.join(currentDirectory,'Doc','mdanse_logo.png'),os.path.join(sphinxDir,'_static'))
            shutil.copy(os.path.join(currentDirectory,'Doc','layout.html'),os.path.join(sphinxDir,'_templates'))
     
            # The directory where the rst files are located.
            self.source_dir = sphinxDir
            # The directory where the conf.py file is located.
            self.config_dir = self.source_dir
     
            # The directory where the documentation will be built
            self.build_dir = os.path.join(buildDir,'MDANSE','Doc',self.doctype)
            
            if isinstance(self.builder,str):
                builders = [self.builder]
            else:
                builders = self.builder

            for builder in builders:
                self.builder_target_dir = os.path.join(self.build_dir, builder)
                sphinx.setup_command.BuildDoc.finalize_options(self)
                sphinx.setup_command.BuildDoc.run(self)
                 
            sys.path.pop(0)
            
    class mdanse_build_help(mdanse_build_doc):
        
        doctype = 'help'

    class mdanse_build_api(mdanse_build_doc):
        
        doctype = 'api'
                        
#################################
# Extensions section
#################################

if 'linux' in sys.platform:
    (opt,) = get_config_vars('OPT')
    os.environ['OPT'] = " ".join(flag for flag in opt.split() if flag != '-Wstrict-prototypes')

CMDCLASS = {}

if sphinx:
    CMDCLASS['build_api'] = mdanse_build_api
    CMDCLASS['build_help'] = mdanse_build_help
             
#################################
# The setup section
#################################

setup (name             = "MDANSE GUI",
       packages=find_packages("Src"),
       package_dir={"": "Src"},
       platforms        = ['Unix','Windows'],
       cmdclass         = CMDCLASS,
    )



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtCore import (
    QPoint,
    Signal,
)
from qtpy.QtWidgets import (
    QToolButton,
    QMenu,
)


class LoaderButton(QToolButton):
    """Subclassed from QToolButton, this object shows the name of a
    chemical element, and creates a pop-up menu giving access to information
    about isotopes when clicked.
    """

    load_hdf = Signal()
    start_converter = Signal(str)

    def __init__(self, *args, caption="Load", backend=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.setText(caption)
        # self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)

        self.converter_source = backend

        self.clicked.connect(self.altContextMenu)

    # def enterEvent(self, a0: QEnterEvent) -> None:
    #     self.atom_info.emit(self.info)
    #     return super().enterEvent(a0)

    def populateMenu(self, menu: QMenu):
        hdf_action = menu.addAction("Load HDF5")
        hdf_action.triggered.connect(self.load_hdf)
        menu.addSeparator()
        for cjob in self.converter_source.backend.getConverters():
            menu.addAction("Convert using " + str(cjob))

    def altContextMenu(self):
        menu = QMenu()
        self.populateMenu(menu)
        res = menu.exec_(self.mapToGlobal(QPoint(10, 10)))
        if res is not None and not "HDF5" in res.text():
            self.start_converter.emit(res.text())

    def contextMenuEvent(self, event):
        menu = QMenu()
        self.populateMenu(menu)
        res = menu.exec_(event.globalPos())
        if res is not None and not "HDF5" in res.text():
            self.start_converter.emit(res.text())



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtCore import QObject, Signal, QMutex

from MDANSE.Framework.Status import Status


class JobCommunicator(QObject):
    target = Signal(int)
    progress = Signal(int)
    finished = Signal(bool)
    oscillate = Signal()


class JobStatusQt(Status):
    def __init__(self, *args, **kwargs):
        qparent = kwargs.pop("parent", None)
        super().__init__()
        self._state = {}  # for compatibility with JobStatus
        self._progress_meter = 0
        self._mutex = QMutex()
        self._communicator = JobCommunicator(parent=qparent)

    @property
    def state(self):
        return self._state

    def finish_status(self):
        self._communicator.finished.emit(True)

    def start_status(self):
        try:
            temp = int(self._nSteps)
        except:
            self._communicator.oscillate.emit()
        else:
            self._communicator.target.emit(temp)

    def stop_status(self):
        self._communicator.finished.emit(False)

    def update_status(self):
        self._mutex.lock()
        self._progress_meter += 1
        temp = int(self._progress_meter)
        self._mutex.unlock()
        self._communicator.progress.emit(temp)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import os
import yaml


from MDANSE.MLogging import LOG


def load_database(database_path):
    """Load in memory the database YAML file"""

    # Load the chemical elements database
    with open(database_path, "r") as fin:
        try:
            database = yaml.safe_load(fin)
        except yaml.YAMLError as exc:
            LOG.error(exc)

    return database


_libdir, _ = os.path.split(__file__)
CHEMICAL_ELEMENTS = load_database(os.path.join(_libdir, "chemical_elements.yml"))
STANDARD_RESIDUES = load_database(os.path.join(_libdir, "residues.yml"))



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from qtpy.QtWidgets import QWidget, QDockWidget, QScrollArea


class WidgetGenerator:
    """The GUI elements will typically belong to one of the few
    possible categories. The generator will make it easier
    to wrap them in some common elements to make them
    dockable, scrollable, etc.
    """

    def wrapWidget(
        self,
        *args,
        cls: QWidget = None,
        parent: QWidget = None,
        name: str = "",
        scrollable: bool = False,
        dockable: bool = False,
        acceptsDrops: bool = True,
        **kwargs,
    ):
        """Takes the constructor for any QWidget
        and creates additional widgets around it
        to enable docking, scrolling, etc.

        Keyword Arguments:
            cls -- the PyQt widget to be created (default: {None})
            parent -- parent QObject for the widget (default: {None})
            name -- name, if needed (default: {""})
            scrollable -- if yes, create a QScrollArea (default: {False})
            dockable -- if yes, create a QDockWidget (default: {False})

        Returns:
            base, instance
            base is the object to be placed in the GUI layout
            instance is the requested widget, which can now
            be configured and connected.
        """
        next_parent = parent
        base = None
        if dockable:
            docker = QDockWidget(name, parent=parent)
            docker.setObjectName(name + "_dockWidget")
            next_parent = docker
            if base is None:
                base = docker
        if scrollable:
            scroller = QScrollArea(next_parent)
            next_parent = scroller
            if base is None:
                base = scroller
        instance = cls(*args, parent=next_parent, **kwargs)
        if acceptsDrops:
            instance.setAcceptDrops(True)
        centre = instance
        if base is None:
            base = instance
        if scrollable:
            scroller.setWidget(centre)
            centre = scroller
        if dockable:
            docker.setWidget(centre)
        return base, instance
